@@ -43,7 +43,9 @@ typedef struct {
 
 typedef struct {
 	GFile			*image_file;
+	GFileMonitor		*image_monitor;
 	guint64			 image_file_size;
+	gboolean		 rename_labels;
 	GPtrArray		*devices;
 	GMutex			 devices_mutex;
 	GSettings		*settings;
@@ -61,6 +63,7 @@ typedef struct {
 	GMutex			 idle_id_mutex;
 	GtkWidget		*switch_verify;
 	GtkWidget		*switch_blank;
+	GtkWidget		*switch_probe;
 } GmwPrivate;
 
 /**
@@ -222,6 +225,120 @@ gmw_device_list_sort (GmwPrivate *priv)
 	}
 }
 
+typedef struct {
+	GmwPrivate	*priv;
+	GmwDevice 	*device;
+	GtkEntry	*entry;
+	GtkDialog	*dialog;
+} GmwRenumberHelper;
+
+static void gmw_refresh_ui (GmwPrivate *priv);
+
+/**
+ * gmw_main_show_quirks:
+ **/
+static void
+gmw_main_show_quirks (GmwPrivate *priv)
+{
+	GmwDevice *device;
+	guint i;
+
+	g_print ("\t/*     hub      hub-port  parent-hub     child-device"
+		 "   chps  dprt  chn  labl */\n");
+	for (i = 0; i < priv->devices->len; i++) {
+		_cleanup_free_ gchar *tmp = NULL;
+		device = g_ptr_array_index (priv->devices, i);
+		tmp = gmw_device_get_quirk_string (device);
+		g_print ("\t%s\n", tmp);
+	}
+}
+
+/**
+ * gmw_main_rename_response_cb:
+ **/
+static void
+gmw_main_rename_response_cb (GtkDialog *dialog,
+			     GtkResponseType response_id,
+			     GmwRenumberHelper *helper)
+{
+	if (response_id == GTK_RESPONSE_OK) {
+		const gchar *tmp;
+		tmp = gtk_entry_get_text (helper->entry);
+		if (tmp != NULL && tmp[0] != '\0') {
+			g_debug ("renaming %s to %s",
+				 gmw_device_get_hub_id (helper->device), tmp);
+			gmw_device_set_hub_label (helper->device, tmp);
+			gmw_device_list_sort (helper->priv);
+			gmw_main_show_quirks (helper->priv);
+		}
+	}
+	gmw_refresh_ui (helper->priv);
+	g_object_unref (helper->device);
+	g_free (helper);
+	gtk_widget_destroy (GTK_WIDGET (dialog));
+}
+
+/**
+ * gmw_main_rename_labels_cb:
+ **/
+static void
+gmw_main_rename_activate_cb (GtkEntry *entry, GmwRenumberHelper *helper)
+{
+	gtk_dialog_response (helper->dialog, GTK_RESPONSE_OK);
+}
+
+/**
+ * gmw_main_rename_labels_cb:
+ **/
+static gboolean
+gmw_main_rename_labels_cb (GtkLabel *label, const gchar *uri, GmwPrivate *priv)
+{
+	GmwDevice *device;
+	GmwRenumberHelper *helper;
+	GtkWidget *w;
+	const gchar *tmp;
+	guint idx;
+
+	/* convert the URI into a device */
+	idx = atoi (uri);
+	device = g_ptr_array_index (priv->devices, idx);
+	if (device == NULL)
+		return FALSE;
+
+	/* create warning dialog */
+	w = GTK_WIDGET (gtk_builder_get_object (priv->builder, "dialog_main"));
+	w = gtk_message_dialog_new (GTK_WINDOW (w),
+				    GTK_DIALOG_DESTROY_WITH_PARENT |
+				    GTK_DIALOG_MODAL |
+				    GTK_DIALOG_USE_HEADER_BAR,
+				    GTK_MESSAGE_INFO,
+				    GTK_BUTTONS_OK_CANCEL,
+				    /* TRANSLATORS: window title renaming labels */
+				    "%s", _("New hub label"));
+
+	/* helper object */
+	helper = g_new0 (GmwRenumberHelper, 1);
+	helper->priv = priv;
+	helper->device = g_object_ref (device);
+	helper->dialog = GTK_DIALOG (w);
+	g_signal_connect (helper->dialog, "response",
+			  G_CALLBACK (gmw_main_rename_response_cb), helper);
+	helper->entry = GTK_ENTRY (gtk_entry_new ());
+	g_signal_connect (helper->entry, "activate",
+			  G_CALLBACK (gmw_main_rename_activate_cb), helper);
+
+	/* add existing hub label */
+	gtk_widget_set_visible (GTK_WIDGET (helper->entry), TRUE);
+	w = gtk_message_dialog_get_message_area (GTK_MESSAGE_DIALOG (helper->dialog));
+	gtk_box_pack_end (GTK_BOX (w), GTK_WIDGET (helper->entry), FALSE, FALSE, 0);
+	tmp = gmw_device_get_hub_label (device);
+	if (tmp != NULL)
+		gtk_entry_set_text (helper->entry, tmp);
+
+	gtk_window_present (GTK_WINDOW (helper->dialog));
+	return TRUE;
+}
+
 /**
  * gmw_refresh_ui:
  **/
@@ -242,6 +359,7 @@ gmw_refresh_ui (GmwPrivate *priv)
 	/* add new children */
 	for (i = 0; i < priv->devices->len; i++) {
 		_cleanup_free_ gchar *label_markup = NULL;
+		_cleanup_free_ gchar *label = NULL;
 		_cleanup_free_ gchar *title = NULL;
 		guint row = i % max_devices_per_column;
 		guint col = (i / max_devices_per_column) * 4;
@@ -250,16 +368,29 @@ gmw_refresh_ui (GmwPrivate *priv)
 
 		/* add label */
 		w = gtk_label_new (NULL);
+		g_signal_connect (w, "activate-link",
+				  G_CALLBACK (gmw_main_rename_labels_cb), priv);
 		if (col > 0)
 			gtk_widget_set_margin_start (w, 30);
 		if (gmw_device_get_hub_label (device) != NULL) {
-			label_markup = g_strdup_printf ("<tt><small>%s [%s]</small></tt>",
-							gmw_device_get_hub_id (device),
-							gmw_device_get_hub_label (device));
+			if (priv->rename_labels) {
+				label = g_strdup_printf ("%s [%s<a href=\"%i\">?</a>]",
+							 gmw_device_get_hub_id (device),
+							 gmw_device_get_hub_label (device), i);
+			} else {
+				label = g_strdup_printf ("%s [%s]",
+							 gmw_device_get_hub_id (device),
+							 gmw_device_get_hub_label (device));
+			}
 		} else {
-			label_markup = g_strdup_printf ("<tt><small>%s</small></tt>",
-							gmw_device_get_hub_id (device));
+			if (priv->rename_labels) {
+				label = g_strdup_printf ("%s [<a href=\"%i\">?</a>]",
+							 gmw_device_get_hub_id (device), i);
+			} else {
+				label = g_strdup (gmw_device_get_hub_id (device));
+			}
 		}
+		label_markup = g_strdup_printf ("<tt><small>%s</small></tt>", label);
 		gtk_label_set_markup (GTK_LABEL (w), label_markup);
 		g_object_set (w, "xalign", 1.f, NULL);
 		gtk_grid_attach (GTK_GRID (grid), w, col + 0, row, 1, 1);
@@ -276,8 +407,12 @@ gmw_refresh_ui (GmwPrivate *priv)
 			w = gtk_progress_bar_new ();
 			gtk_widget_set_valign (w, GTK_ALIGN_CENTER);
 			gtk_grid_attach (GTK_GRID (grid), w, col + 2, row, 1, 1);
-			gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (w),
-						       gmw_device_get_complete (device));
+			if (gmw_device_get_complete (device) <= 100.f) {
+				gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (w),
+							       gmw_device_get_complete (device));
+			} else {
+				gtk_progress_bar_pulse (GTK_PROGRESS_BAR (w));
+			}
 		}
 
 		/* add optional status text */
@@ -665,8 +800,8 @@ gmw_device_verify (GmwPrivate *priv,
 		/* compare */
 		if (memcmp (buffer_src, buffer_dest, bytes_to_read) != 0) {
 			g_set_error (error, 1, 0,
-				     "Failed to verify at %" G_GUINT64_FORMAT,
-				     bytes_completed);
+				     "Failed to verify at %" G_GUINT64_FORMAT "Mb",
+				     bytes_completed / (1024 * 1024));
 			goto out;
 		}
 
@@ -743,6 +878,35 @@ gmw_refresh_titlebar_idle_cb (gpointer user_data)
 	return G_SOURCE_REMOVE;
 }
 
+/**
+ * gmw_block_device_probe:
+ **/
+static gboolean
+gmw_block_device_probe (const gchar *block_dev, GError **error)
+{
+	gboolean ret;
+	gint exit_status = 0;
+	_cleanup_free_ gchar *standard_output = NULL;
+	const gchar *argv[] = { "/usr/bin/pkexec",
+				"/usr/bin/gnome-multi-writer-probe",
+				block_dev,
+				NULL };
+	const gchar *envp[] = { NULL };
+
+	ret = g_spawn_sync (NULL, (gchar **) argv, (gchar **) envp,
+			    G_SPAWN_STDERR_TO_DEV_NULL,
+			    NULL, NULL,
+			    &standard_output, NULL, &exit_status, error);
+	if (!ret)
+		return FALSE;
+	g_strdelimit (standard_output, "\n", '\0');
+	if (exit_status != 0) {
+		g_set_error_literal (error, 1, 0, standard_output);
+		return FALSE;
+	}
+	return TRUE;
+}
+
 /**
  * gmw_copy_thread_cb:
  **/
@@ -759,6 +923,17 @@ gmw_copy_thread_cb (gpointer data, gpointer user_data)
 	if (!g_settings_get_boolean (priv->settings, "enable-verify"))
 		gmw_device_set_write_alloc (device, 1.f);
 
+	/* probe the devices */
+	if (g_settings_get_boolean (priv->settings, "enable-probe")) {
+		const gchar *block_dev;
+		gmw_device_set_state (device, GMW_DEVICE_STATE_VERIFY);
+		block_dev = gmw_device_get_block_path (device);
+		if (!gmw_block_device_probe (block_dev, &error)) {
+			gmw_device_set_error (device, error);
+			goto out;
+		}
+	}
+
 	/* open input stream */
 	image_stream = (GInputStream *) g_file_read (priv->image_file, NULL, &error);
 	if (image_stream == NULL) {
@@ -788,7 +963,8 @@ out:
 	gmw_refresh_in_idle (priv);
 	g_idle_add (gmw_refresh_titlebar_idle_cb, priv);
 	g_timeout_add_seconds (2, gmw_refresh_in_idle_cb, priv);
-	g_input_stream_close (G_INPUT_STREAM (image_stream), NULL, NULL);
+	if (image_stream != NULL)
+		g_input_stream_close (G_INPUT_STREAM (image_stream), NULL, NULL);
 	gmw_copy_done (priv);
 }
 
@@ -809,6 +985,19 @@ gmw_update_title (GmwPrivate *priv)
 	}
 }
 
+/**
+ * gmw_set_image_file_changed_cb:
+ **/
+static void
+gmw_set_image_file_changed_cb (GFileMonitor *monitor,
+			       GFile *file,
+			       GFile *other_file,
+			       GFileMonitorEvent event_type,
+			       GmwPrivate *priv)
+{
+	g_debug ("ISO file changed");
+}
+
 /**
  * gmw_set_image_filename:
  **/
@@ -820,7 +1009,19 @@ gmw_set_image_filename (GmwPrivate *priv, const gchar *filename)
 
 	if (priv->image_file != NULL)
 		g_object_unref (priv->image_file);
+	if (priv->image_monitor != NULL)
+		g_object_unref (priv->image_monitor);
 	priv->image_file = g_file_new_for_path (filename);
+	priv->image_monitor = g_file_monitor (priv->image_file,
+					      G_FILE_MONITOR_NONE,
+					      NULL, &error);
+	if (priv->image_monitor == NULL) {
+		/* TRANSLATORS: we couldn't open the ISO file the user chose */
+		gmw_error_dialog (priv, _("Failed to open"), error->message);
+		return;
+	}
+	g_signal_connect (priv->image_monitor, "changed",
+			  G_CALLBACK (gmw_set_image_file_changed_cb), priv);
 
 	/* get the size of the ISO file */
 	info = g_file_query_info (priv->image_file,
@@ -1161,18 +1362,14 @@ gmw_about_activated_cb (GSimpleAction *action, GVariant *parameter, gpointer use
 {
 	GmwPrivate *priv = (GmwPrivate *) user_data;
 	GList *windows;
-	GtkIconTheme *icon_theme;
 	GtkWindow *parent = NULL;
 	const gchar *authors[] = { "Richard Hughes", NULL };
 	const gchar *copyright = "Copyright \xc2\xa9 2014-2015 Richard Hughes";
-	_cleanup_object_unref_ GdkPixbuf *logo = NULL;
 
 	windows = gtk_application_get_windows (GTK_APPLICATION (priv->application));
 	if (windows)
 		parent = windows->data;
 
-	icon_theme = gtk_icon_theme_get_default ();
-	logo = gtk_icon_theme_load_icon (icon_theme, "gnome-multi-writer", 256, 0, NULL);
 	gtk_show_about_dialog (parent,
 			       /* TRANSLATORS: the title of the about window */
 			       "title", _("About GNOME MultiWriter"),
@@ -1183,7 +1380,7 @@ gmw_about_activated_cb (GSimpleAction *action, GVariant *parameter, gpointer use
 			       "comments", _("Write an ISO file to multiple USB devices at once"),
 			       "copyright", copyright,
 			       "license-type", GTK_LICENSE_GPL_2_0,
-			       "logo", logo,
+			       "logo-icon-name", "gnome-multi-writer",
 			       /* TRANSLATORS: you can put your name here :) */
 			       "translator-credits", _("translator-credits"),
 			       "version", VERSION,
@@ -1224,6 +1421,9 @@ gmw_settings_clicked_cb (GtkWidget *widget, GmwPrivate *priv)
 	if (gtk_widget_get_parent (priv->switch_blank) != NULL)
 		g_object_ref (priv->switch_blank);
 	gtk_widget_unparent (priv->switch_blank);
+	if (gtk_widget_get_parent (priv->switch_probe) != NULL)
+		g_object_ref (priv->switch_probe);
+	gtk_widget_unparent (priv->switch_probe);
 
 	/* show settings */
 	pop = gtk_popover_new (widget);
@@ -1244,6 +1444,12 @@ gmw_settings_clicked_cb (GtkWidget *widget, GmwPrivate *priv)
 			 gtk_label_new (_("Wipe")),
 			 0, 1, 1, 1);
 	gtk_grid_attach (GTK_GRID (box), priv->switch_blank, 1, 1, 1, 1);
+	gtk_grid_attach (GTK_GRID (box),
+			 /* TRANSLATORS: a switch label: we check the device
+			  * is actually the size it says it is */
+			 gtk_label_new (_("Probe")),
+			 0, 2, 1, 1);
+	gtk_grid_attach (GTK_GRID (box), priv->switch_probe, 1, 2, 1, 1);
 	gtk_container_add (GTK_CONTAINER (pop), box);
 	gtk_widget_show_all (pop);
 }
@@ -1296,13 +1502,9 @@ gmw_startup_cb (GApplication *application, GmwPrivate *priv)
 
 	/* setup USB image */
 	w = GTK_WIDGET (gtk_builder_get_object (priv->builder, "image_usb"));
-	pixbuf = gdk_pixbuf_new_from_resource_at_scale ("/org/gnome/MultiWriter/usb.svg",
-							-1, 48, TRUE, &error);
-	if (pixbuf == NULL) {
-		g_warning ("failed to load usb.svg: %s", error->message);
-		return;
-	}
-	gtk_image_set_from_pixbuf (GTK_IMAGE (w), pixbuf);
+	gtk_image_set_from_resource (GTK_IMAGE (w),
+				     "/org/gnome/MultiWriter/usb-symbolic.svg");
+	gtk_image_set_pixel_size (GTK_IMAGE (w), 48);
 
 	/* get previously loaded image */
 	filename = g_settings_get_string (priv->settings, "filename");
@@ -1408,11 +1610,6 @@ gmw_udisks_object_add (GmwPrivate *priv, GDBusObject *dbus_object)
 		g_debug ("%s has no org.freedesktop.UDisks2.Partition", object_path);
 		return FALSE;
 	}
-	iface_fs = g_dbus_object_get_interface (dbus_object, "org.freedesktop.UDisks2.Filesystem");
-	if (iface_fs != NULL) {
-		g_debug ("%s has no org.freedesktop.UDisks2.Filesystem", object_path);
-		return FALSE;
-	}
 
 	/* get the block device */
 	udisks_object = udisks_client_get_object (priv->udisks_client, object_path);
@@ -1429,6 +1626,12 @@ gmw_udisks_object_add (GmwPrivate *priv, GDBusObject *dbus_object)
 		return FALSE;
 	}
 
+	/* ignore SD cards */
+	if (g_str_has_prefix (block_path, "/dev/mmcblk")) {
+		g_debug ("%s is not a USB device", block_path);
+		return FALSE;
+	}
+
 	/* ignore small or large devices */
 	device_size = udisks_block_get_size (udisks_block) / (1000 * 1000);
 	if (device_size < 1000) {
@@ -1436,7 +1639,7 @@ gmw_udisks_object_add (GmwPrivate *priv, GDBusObject *dbus_object)
 			 block_path, (guint) device_size);
 		return FALSE;
 	}
-	if (device_size > 1000 * 16) {
+	if (device_size > 1024 * 32) {
 		g_debug ("%s is too large [%u]",
 			 block_path, (guint) device_size);
 		return FALSE;
@@ -1605,9 +1808,13 @@ main (int argc, char **argv)
 	GmwPrivate *priv = NULL;
 	GOptionContext *context;
 	gboolean verbose = FALSE;
+	gboolean rename_labels = FALSE;
 	int status = EXIT_SUCCESS;
 	_cleanup_error_free_ GError *error = NULL;
 	const GOptionEntry options[] = {
+		{ "rename-labels", '\0', 0, G_OPTION_ARG_NONE, &rename_labels,
+			/* TRANSLATORS: command line option */
+			_("Allow renaming the labels on hubs"), NULL },
 		{ "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose,
 			/* TRANSLATORS: command line option */
 			_("Show extra debugging information"), NULL },
@@ -1635,6 +1842,7 @@ main (int argc, char **argv)
 	}
 
 	priv = g_new0 (GmwPrivate, 1);
+	priv->rename_labels = rename_labels;
 	g_mutex_init (&priv->thread_pool_mutex);
 	g_mutex_init (&priv->devices_mutex);
 	g_mutex_init (&priv->idle_id_mutex);
@@ -1676,12 +1884,16 @@ main (int argc, char **argv)
 	/* keep these local as they get reparented to the popover */
 	priv->switch_verify = gtk_switch_new ();
 	priv->switch_blank = gtk_switch_new ();
+	priv->switch_probe = gtk_switch_new ();
 	g_settings_bind (priv->settings, "enable-verify",
 			 priv->switch_verify, "active",
 			 G_SETTINGS_BIND_DEFAULT);
 	g_settings_bind (priv->settings, "blank-drive",
 			 priv->switch_blank, "active",
 			 G_SETTINGS_BIND_DEFAULT);
+	g_settings_bind (priv->settings, "enable-probe",
+			 priv->switch_probe, "active",
+			 G_SETTINGS_BIND_DEFAULT);
 
 	/* wait */
 	status = g_application_run (G_APPLICATION (priv->application), argc, argv);
@@ -1698,6 +1910,8 @@ out:
 			g_object_unref (priv->usb_ctx);
 		if (priv->image_file != NULL)
 			g_object_unref (priv->image_file);
+		if (priv->image_monitor != NULL)
+			g_object_unref (priv->image_monitor);
 		if (priv->cancellable != NULL)
 			g_object_unref (priv->cancellable);
 		if (priv->application != NULL)
